package com.atl.afaultdetection.utils;

import com.atl.mcod.utils.Data;
import com.atl.mcod.utils.Constants;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.time.Duration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import okhttp3.OkHttpClient;
import org.influxdb.InfluxDB;
import org.influxdb.InfluxDBFactory;
import org.influxdb.dto.Point;
import org.influxdb.dto.Query;
import org.influxdb.dto.QueryResult;

/**
 *
 * @author thanasis
 */
public class InfluxDBHandler {

    private InfluxDB influxDB;
    private String database;
    private String resultsDatabase;
    private Map<String, Data> lastReportedMap; //to avoid multiple reports

//    public InfluxDBHandler(String host, String port, String username, String password, String database) {
//        OkHttpClient.Builder builder = new OkHttpClient.Builder()
//                .readTimeout(100, TimeUnit.SECONDS).connectTimeout(1000, TimeUnit.SECONDS);
//        influxDB = InfluxDBFactory.connect("https://" + host + ":" + port, username, password, builder);
//        influxDB.setDatabase(database);
//        this.database = database;
//        this.lastReportedMap = new HashMap<>();
//    }

    public InfluxDBHandler(String host, String port, String database, String resutlsDatabase, String username, String password, boolean ssl) {
        try {
            OkHttpClient.Builder builder = null;
            String http = "http://";
            if(ssl){
                // Create a trust manager that does not validate certificate chains
                final TrustManager[] trustAllCerts = new TrustManager[]{
                    new X509TrustManager() {
                        @Override
                        public void checkClientTrusted(java.security.cert.X509Certificate[] chain, String authType) throws CertificateException {
                        }

                        @Override
                        public void checkServerTrusted(java.security.cert.X509Certificate[] chain, String authType) throws CertificateException {
                        }

                        @Override
                        public java.security.cert.X509Certificate[] getAcceptedIssuers() {
                            return new java.security.cert.X509Certificate[]{};
                        }
                    }
                };

                // Install the all-trusting trust manager
                SSLContext sc = SSLContext.getInstance("SSL");
                sc.init(null, trustAllCerts, new java.security.SecureRandom());

                builder = new OkHttpClient.Builder()
                        .sslSocketFactory(sc.getSocketFactory(), (X509TrustManager) trustAllCerts[0]).readTimeout(100, TimeUnit.SECONDS).connectTimeout(1000, TimeUnit.SECONDS);
                builder.hostnameVerifier(new HostnameVerifier() {
                    @Override
                    public boolean verify(String hostname, SSLSession session) {
                        return true;
                    }
                });
                http = "https://";
            } else {
                builder = new OkHttpClient.Builder()
                    .readTimeout(100, TimeUnit.SECONDS).connectTimeout(1000, TimeUnit.SECONDS);
                http = "http://";
            }
            if(username == null || username.equals("")){
                influxDB = InfluxDBFactory.connect(http + host + ":" + port, builder);
            } else {
                influxDB = InfluxDBFactory.connect(http + host + ":" + port, username, password, builder);
            }
            influxDB.setDatabase(database);
            this.database = database;
            this.resultsDatabase = resutlsDatabase;
            
            if(!dbExists(this.resultsDatabase)){
                influxDB.query(new Query("create database " + this.resultsDatabase + " with name \"autogen\"",resutlsDatabase));
            }
            
            this.lastReportedMap = new HashMap<>();
        } catch (NoSuchAlgorithmException ex) {
            Logger.getLogger(InfluxDBHandler.class.getName()).log(Level.SEVERE, null, ex);
        } catch (KeyManagementException ex) {
            Logger.getLogger(InfluxDBHandler.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public InfluxDB getInfluxDBHandler() {
        return influxDB;
    }

    public QueryResult getMeasurements(String measurementName, long seconds, boolean excludeZeroValues){
        if(influxDB != null){
            influxDB.setDatabase(this.database);
            Query query = new Query("SELECT mean(value) FROM "+measurementName+" WHERE time > now() - "+seconds+"s and time <= now() "+(excludeZeroValues?"and value != 0":"")+" group by time(1s) fill(linear)", database);
            QueryResult results = influxDB.query(query);
            return results;
        } else {
            System.err.print("No DB connection");
        }
        return new QueryResult();
    }
    //for debugging
    public QueryResult getMeasurementsDebug(String measurementName, long seconds, boolean excludeZeroValues) {
        if (influxDB != null) {
            influxDB.setDatabase(this.database);
//            Query query = new Query("SELECT mean(value) FROM " + measurementName + " WHERE time > '2018-01-25T00:00:01.004Z' + " + (seconds - Constants.slide) + "s and time <= '2018-01-25T00:00:01.004Z' + " + seconds + "s " + (excludeZeroValues ? "and value != 0" : "") + " group by time(1s) fill(linear)", database);
            //Query query = new Query("SELECT mean(value) FROM " + measurementName + " WHERE time > '2018-07-05T00:00:01.004Z' + " + (seconds - Constants.slide) + "s and time <= '2018-07-05T00:00:01.004Z' + " + seconds + "s " + (excludeZeroValues ? "and value != 0" : "") + " group by time(1s) fill(linear)", database);
            //FOR AE TESTING
            Query query = new Query("SELECT value AS \"mean\" FROM " + measurementName + " WHERE time > '2018-08-19T00:00:01.004Z' + " + (seconds - Constants.slide) + "s and time <= '2018-08-19T00:00:01.004Z' + " + seconds + "s " /*+ (excludeZeroValues ? "and value != 0" : "") + ""*/, database);
            //FOR THICK TESTING
            //Query query = new Query("SELECT value AS \"mean\" FROM " + measurementName + " WHERE time > '2018-10-14T00:00:01.004Z' + " + (seconds - Constants.slide) + "s and time <= '2018-10-14T00:00:01.004Z' + " + seconds + "s " /*+ (excludeZeroValues ? "and value != 0" : "") + ""*/, database);
            QueryResult results = influxDB.query(query);
            return results;
        } else {
            System.err.print("No DB connection");
        }
        return new QueryResult();
    }

    public Double getMinValue(String measurementName, boolean excludeZeroValues, boolean excludeNegativeValues) {
        if (influxDB != null) {
            influxDB.setDatabase(this.database);
            Query query = new Query("SELECT min(value) FROM " + measurementName + (excludeNegativeValues && excludeZeroValues ? " where value > 0" : excludeZeroValues ? " where value != 0" : excludeNegativeValues ? " where value >= 0" : ""), database);
            QueryResult results = influxDB.query(query);
            return (Double) results.getResults().get(0).getSeries().get(0).getValues().get(0).get(1);
        } else {
            System.err.print("No DB connection");
        }
        return null;
    }

    public Double getMaxValue(String measurementName) {
        if (influxDB != null) {
            influxDB.setDatabase(this.database);
            Query query = new Query("SELECT max(value) FROM " + measurementName, database);
            QueryResult results = influxDB.query(query);
            return (Double) results.getResults().get(0).getSeries().get(0).getValues().get(0).get(1);
        } else {
            System.err.print("No DB connection");
        }
        return null;
    }

    public void reportOutlier(Data outlier, String label, String ruleMeas) {
        influxDB.setDatabase(this.resultsDatabase);
        //don't report outliers with distance lower than 5mins
//        if (lastReportedMap.containsKey(label)) {
//            if (Duration.between(lastReportedMap.get(label).getActualTime(), outlier.getActualTime()).toMinutes() < 5 ) {
//                return;
//            }
//        }
        if(Constants.debug || Constants.verbose){
            //System.out.println(label + " " + outlier.getActualTime().toString());
        }
        
        influxDB.setRetentionPolicy("autogen");
        influxDB.write(Point.measurement(Constants.resutlsDBTable)
                .time(outlier.getActualTime().toEpochMilli(), TimeUnit.MILLISECONDS)
                //.addField("report_time", System.currentTimeMillis())
                .addField("report_time", outlier.getOutlier_start_time())
                .addField("reason", "")
                .tag("sensor", ruleMeas.isEmpty()?Constants.sensors:ruleMeas)
                .tag("seen", "")
                .tag("submitted", "")
                .tag("machine", "A-ERWAERM1-PRESSE")
                .tag("label", label)
                .build());
        lastReportedMap.put(label, outlier);
    }
    
    private boolean dbExists(String database){
        boolean exists = false;
        QueryResult query = influxDB.query(new Query("show databases",database));
        for (List<Object> v : query.getResults().get(0).getSeries().get(0).getValues()){
            if(v.get(0).equals(database)){
               return true;   
            }
        }
        return exists;
    }
}
