package com.atl.smartmaintenance;

import com.atl.smartmaintenance.faultdetection.FaultDetectionManager;
import com.atl.afaultdetection.utils.Measurement;
import com.atl.mcod.utils.Constants;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author thanasis
 */
public class AFaultDetection {
    
    public static void main(String[] args) {
        try {
            readArguments(args);
            FaultDetectionManager fdm = new FaultDetectionManager();
            Constants.formSensors();
            fdm.detectOutliers(Constants.measMap);
        } catch (InterruptedException ex) {
            Logger.getLogger(AFaultDetection.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    public static void readArguments(String[] args) {
        for (int i = 0; i < args.length; i++) {

            //check if arg starts with --
            String arg = args[i];
            if (arg.indexOf("--") == 0) {
                switch (arg) {
                    case "--R":
                        Constants.R = Double.valueOf(args[i + 1]);
                        break;
                    case "--W":
                        Constants.W = Integer.valueOf(args[i + 1]);
                        break;
                    case "--k":
                        Constants.k = Integer.valueOf(args[i + 1]);
                        break;
                    case "--slide":
                        Constants.slide = Integer.valueOf(args[i + 1]);
                        break;
                    case "--R_k":
                        Constants.R_k = Double.valueOf(args[i + 1]);
                        break;
                    case "--k_k":
                        Constants.k_k = Integer.valueOf(args[i + 1]);
                        break;
                    case "--meas":
                        for(String meas : args[i+1].split(",")){
                            Constants.measMap.put(Constants.measShortNameMap.get(meas), new Measurement("\""+Constants.measShortNameMap.get(meas)+"\""));
                        }
                        break;
                    case "--algorithms":
                        Set<String> algorithms = new HashSet<>();
                        for (String algo : args[i + 1].split(",")) {
                            algorithms.add(algo);
                        }
                        Constants.algorithms = algorithms;
                        break;
                    case "--stamperLockedPosition":
                        Constants.stamperLockedPosition = Double.valueOf(args[i + 1]);
                        break;
                    case "--stabilizationPeriod":
                        Constants.stabilizationPeriod = Integer.valueOf(args[i + 1]);
                        break;
                    case "--resultsDB":
                        Constants.resutlsDB = args[i + 1];
                        break;
                    case "--resultsDBTable":
                        Constants.resutlsDBTable = args[i + 1];
                        break;
                    case "--maxRule":
                        Constants.maxRule = args[i + 1];
                        break;
                    case "--minRule":
                        Constants.minRule = args[i + 1];
                        break;
                    case "--maxRule_k":
                        Constants.maxRule_k = args[i + 1];
                        break;
                    case "--minRule_k":
                        Constants.minRule_k = args[i + 1];
                        break;
                    case "--outlierLife":
                        Constants.outlierLife = Double.valueOf(args[i + 1]);
                        break;
                    case "--outlierLife_k":
                        Constants.outlierLife_k = Double.valueOf(args[i + 1]);
                        break;                        
                    case "--host":
                        Constants.host = args[i + 1];
                        break;
                    case "--port":
                        Constants.port = args[i + 1];
                        break;
                    case "--username":
                        Constants.username = args[i + 1];
                        break; 
                    case "--password":
                        Constants.password = args[i + 1];
                        break; 
                    case "--database":
                        Constants.database = args[i + 1];
                        break;
                    case "--debug":
                        Constants.debug = Boolean.parseBoolean(args[i + 1]);
                        break;
                    case "--verbose":
                        Constants.verbose = Boolean.parseBoolean(args[i + 1]);
                        break;
                    case "--ssl":
                        Constants.ssl = Boolean.parseBoolean(args[i + 1]);
                        break;
                    case "--help":
                        System.out.println("Usage of aFaultDetection:");
                        System.out.println("\t--host\n"
                                + "\t\tSpecify the influxDB host (hostname or ip address).\n"
                                + "\t--port\n"
                                + "\t\tSpecify the influxDB port.\n"
                                + "\t--username\n"
                                + "\t\tSpecify the influxDB username (optional). You will be prompt for a password if not specified \n"
                                + "\t\twith --passowrd.\n"
                                + "\t--password\n"
                                + "\t\tSpecify the influxDB password (optional).\n"
                                + "\t--database\n"
                                + "\t\tSpecify the influxDB database.\n"
                                + "\t--ssl\n"
                                + "\t\tSpecify whether to use ssl certification (https://) or not. ssl is used without verification.\n"
                                + "\t\t(default: false).\n"
                                + "\t--algorithms mcod,mcodKnowledge,ruleBased,ruleBasedKnowledge\n"
                                + "\t\tSpecify the used algorithms for the fault detection (multiple selections allowed separated with \n"
                                + "\t\tcomma (no spaces)): mcod, mcodKnowledge, ruleBased, ruleBasedKnowledge\n"
                                + "\t--meas oilLevel,lowPressurePumps,highPressurePumps,pressForce,stamperPosition,productNumber,oilTemperature,\n"
                                + "\t       cylinderPressure1,cylinderPressure2,cylinderPressure3,cylinderPressure4\n"
                                + "\t\tSpecify the measurments to monitor (multiple selections allowed separated with comma (no spaces)): \n"
                                + "\t\toilLevel, lowPressurePumps, highPressurePumps, pressForce, stamperPosition, productNumber, \n"
                                + "\t\toilTemperature, cylinderPressure1, cylinderPressure2, cylinderPressure3, cylinderPressure4\n"
                                + "\t--R, --R_k\n"
                                + "\t\tSet the R radius of the mcod and mcodKnowledge algorithms, respectively. The higher the R the less \n"
                                + "\t\tthe outliers (default: R=0.15, R_k=0.08).\n"
                                + "\t--k, --k_k\n"
                                + "\t\tSet the minimum number of points inside radius R (R_k resp.) to form a cluster of inlier points for \n"
                                + "\t\tmcod and mcodKnowledge, respectively. The higher the k the more the outliers (default: k=400, k_k=720).\n"
                                + "\t--W\n"
                                + "\t\tSet the window size for both mcod and mcodKnowledge. Higher window sizes demand more \n"
                                + "\t\tprocessing power (default: 3600).\n"
                                + "\t--slide\n"
                                + "\t\tSet the slide size for both mcod and mcodKnowledge. Defines the movement of the window \n"
                                + "\t\t(i.e. the frequency of data fetching from influxDB) (default: 360).\n"
                                + "\t--resultsDB\n"
                                + "\t\tSpecify the database that the outliers are going to be reported \n"
                                + "\t\t(default: BoostAnmlyDtctn).\n"
                                + "\t--resultsDBTable\n"
                                + "\t\tSpecify the table name of the influxDB that the outliers are going to be reported \n"
                                + "\t\t(default: anomalies2).\n"
                                + "\t--verbose\n"
                                + "\t\tReport the detected outliers also in the stdout \n"
                                + "\t\t(default: false).\n"
                                + "\t--stamperLockedPoisition\n"
                                + "\t\tSet the position of the stamper in meters, where the stamper is considered locked. \n"
                                + "\t\tUsed only in the mcodKnowledge algorithm (default: 2198).\n"
                                + "\t--stabilizationPeriod\n"
                                + "\t\tSet the stabilization period in seconds where the oil is considered stabilized after the stamper is locked. \n"
                                + "\t\tUsed only in the mcodKnowledge algorithm (default: 10).\n"
                                + "\t--outlierLife, --outlierLife_k\n"
                                + "\t\tSet the life of an outlier for mcod and mcodKnowledge, respectively, expressed as percentage of total slides \n"
                                + "\t\tper window (i.e. how many slides should the outlier be active to be reported). The higher the outlierLife the \n"
                                + "\t\tless the reported outliers. (default: outlierLife=0.7, outlierLife_k=0.1).\n"
                                + "\t--minRule --minRule_k oilLevel-<value>|lowPressurePumps-<value>|highPressurePumps-<value>|pressForce-<value>\n"
                                + "\t          |stamperPosition-<value>|productNumber-<value>|oilTemperature-<value>|cylinderPressure1-<value>\n"
                                + "\t          |cylinderPressure2-<value>|cylinderPressure3-<value>|cylinderPressure4-<value>\n"
                                + "\t\tSpecify a rule with minimum threshold e.g. --minRule oilLevel-8330 reports a fault if oilLevel<8330. \n"
                                + "\t\tIt can be used with --maxRule, OR condition is considered between them. --minRule_k is for the ruleBasedKnowledge\n"
                                + "\t--maxRule --maxRule_k oilLevel-<value>|lowPressurePumps-<value>|highPressurePumps-<value>|pressForce-<value>\n"
                                + "\t          |stamperPosition-<value>|productNumber-<value>|oilTemperature-<value>|cylinderPressure1-<value>\n"
                                + "\t          |cylinderPressure2-<value>|cylinderPressure3-<value>|cylinderPressure4-<value>\n"
                                + "\t\tSpecify a rule with maximum threshold e.g. --maxRule oilLevel-9500 reports a fault if oilLevel>9500. \n"
                                + "\t\tIt can be used with --minRule, OR condition is considered between them. --maxRule_k is for the ruleBasedKnowledge");
                        System.out.println("\nExamples:\n");
                        System.out.println("\t# Monitor the oil level with mcodKnowledge and ruleBasedKnoledge with default parameters for \n"
                                + "\tmcodKnowledge and oilLevel<8400 rule for the ruleBasedKnowledge:\n"
                                + "\tjava -jar aFaultDetection-<version>.jar --host localhost --port 8086 --database Axoom1 \n"
                                + "\t--algorithms mcodKnowledge,ruleBasedKnoledge --minRule_k oilLevel-8400\n\n"
                                + "\t# Monitor the oil level and low pressure pumps with mcod seting R=0.5 euclidean distance, k=600 points, \n"
                                + "\tW=5mins (360secs), slide=5secs:\n"
                                + "\tjava -jar aFaultDetection-<version>.jar --host localhost --port 8086 --database Axoom1 --algorithms mcod --R 0.5 --k 600 --W 360 --slide 5\n\n"
                                + "\t# Monitor the oilLevel and lowPressurePums, connect to the influxDB using ssl, write the results to the BoostAnmlyDtctn database,\n "
                                + "\tanomalies2 Measurements/Table, use all the avalable monitoring approaches, providing all the possible parameters:\n"
                                + "\tjava -jar aFaultDetection-<version>.jar --host localhost --port 8086 --database Axoom1 --ssl true --resultsDB BoostAnmlyDtctn\n"
                                + "\t--resultsDBTable anomalies2 --meas oilLevel,lowPressurePumps --algorithms mcod,ruleBased,mcodKnowledge,ruleBasedKnowledge \n"
                                + "\t--minRule oilLevel-9130 --minRule_k oilLevel-9330 --W 3600 --slide 360 --R 0.5 --k 22 --oulierLife 1 --R_k 0.12 --k_k 720 \n"
                                + "\t--outlierLife_k 1 --stamperLockedPosition 2198 --stabilizationPeriod 10 --verbose true");
                        System.exit(0);
                }
            }
        }
    }
    
}
