/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.atl.smartmaintenance.faultdetection;

import com.atl.afaultdetection.utils.InfluxDBHandler;
import com.atl.mcod.MCOD;
import com.atl.afaultdetection.utils.Measurement;
import com.atl.afaultdetection.utils.Utils;
import com.atl.mcod.utils.Constants;
import com.atl.mcod.utils.Data;
import java.time.Instant;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 *
 * @author thanasis
 */
public class FD {
    
    private MCOD mcod;
    private InfluxDBHandler infDBH;
    private double R;
    private int k;
    private int W;
    private int slide;
    private double outlierLife;
    private String ruleMeas;
    private Map<String, Measurement> measMap;
    private boolean stamperPositionExplicitlyAdded; //MCOD with Knowledge needs stamper position to define if stamper is locked
    private boolean productNumberExplicitlyAdded; //MCOD with Knowledge needs product number as for some products (like 3 and 12) the stamper is not clear where it is locked
    private long currentTime;
    
    //Holds the data of each slide
    //MCOD with knowledge is count-based not time based, hence the slideDataList should be processed only if it's size is equal to the slide size
    private List<Data> slideDataList;
    private List<Data> notProcessedDataList;

    public FD(Map<String,Measurement> measMap, InfluxDBHandler infDBH, double R, int k, int W, int slide, double outlierLife, boolean stamperPositionExplicitlyAdded, boolean productNumberExplicitlyAdded) {
        this.R = R;
        this.k = k;
        this.W = W;
        this.slide = slide;
        this.outlierLife = outlierLife;
        this.mcod = new MCOD(R, k, W, slide);
        this.infDBH = infDBH;
        this.stamperPositionExplicitlyAdded = stamperPositionExplicitlyAdded;
        this.productNumberExplicitlyAdded = productNumberExplicitlyAdded;
        this.measMap = measMap;
        if (Constants.algorithms.contains("ruleBased")) {
            this.ruleMeas = Utils.checkRuleBasedParameters(measMap);
        }
        this.slideDataList = new ArrayList<Data>();
        this.notProcessedDataList = new ArrayList<Data>();
        this.currentTime = 0;
    }
    
    public void mcodNoKnowledge(Map<String,Iterator<Map.Entry<Instant,Double>>> slideDataIteratorsMap) {
        
        if(!notProcessedDataList.isEmpty()){
            if(notProcessedDataList.size() <= Constants.slide){
                slideDataList.addAll(notProcessedDataList);
                notProcessedDataList.clear();
                currentTime += slideDataList.size();
            } else {
                currentTime += Constants.slide;
                mcodOutliers(mcod, notProcessedDataList.subList(0, Constants.slide-1), currentTime);
                notProcessedDataList.subList(0, Constants.slide).clear();
                slideDataList.addAll(notProcessedDataList);
                notProcessedDataList.clear();
                currentTime += slideDataList.size();
            }
        }
        
        //loop until one of the iterator runs out of data. 
        //Since we have used group by and fill(linear) in the query it is expected all the iterators to have the same amount of data.
        iteratorsLoop:
        while (Utils.allIteratorsHaveNext(slideDataIteratorsMap)) {
            Instant date = null;//it is used to mark the actual arrival date of the current data point. We have used precision of a second in the query so it is expected all the current points from different measurements to have the same timestamp

            boolean skipValues = false;
            Map<String, Map.Entry<Instant,Double>> valuesMap = new HashMap<>(); //holds the actual <time,point> pair for each different measurment
            for (Map.Entry<String, Iterator<Map.Entry<Instant,Double>>> entry : slideDataIteratorsMap.entrySet()) {
                String key = entry.getKey();
                Map.Entry<Instant,Double> value = entry.getValue().next();

                date = value.getKey();

                //There might be a bug in the InfluxDB java library, as the last returned value is always null.
                //Skip the null values.
                if (value.getValue() == null) {
                    skipValues = true; //continue the outer loop
                    continue iteratorsLoop;
                }

                //InfluxDB does not accept constraints in specific values of multi-measurement queries
                //Hence the exclusion of zero oil level values (which is possibly a sensor fault) should be done manually.
                if (value.getValue() == 0 && key.contains(Constants.measShortNameMap.get("oilLevel"))) {
                    currentTime++;
                    continue iteratorsLoop;
                }

                //Ignore the values that are less than zero which is possibly caused by a sensor problem
                if (value.getValue() < 0) {
                    currentTime++;
                    continue iteratorsLoop;
                }
                valuesMap.put(key, value);
            }
            if(skipValues){
                continue;
            }
            
            //a flag to report data that violate a rule based threshold
            boolean reportDataOutlierRuleBased = false;

            //The implementation of the MCOD expects all the data point values in an array format
            //subtract the explicitly added fields
            double[] d = new double[valuesMap.size()-(stamperPositionExplicitlyAdded && productNumberExplicitlyAdded?2:stamperPositionExplicitlyAdded?1:productNumberExplicitlyAdded?1:0)];
            int i = 0;
            for (String key : valuesMap.keySet()) {
                double value = valuesMap.get(key).getValue();
                if((key.contains(Constants.measShortNameMap.get("stamperPosition")) && stamperPositionExplicitlyAdded) || (key.contains(Constants.measShortNameMap.get("productNumber")) && productNumberExplicitlyAdded)){
                    continue;
                }
                
                if(Constants.algorithms.contains("ruleBased") && key.equals(ruleMeas)){
                    reportDataOutlierRuleBased = Utils.checkRule("ruleBased",key,value,ruleMeas);
                }
                
                Measurement meas = measMap.get(key);
                d[i] = Utils.normalize(value, meas.getMinValue(), meas.getMaxValue());
                i++;
            }

            //Data class is needed by the current MCOD implementation
            Data data = new Data(d);
            data.setActualTime(date);
            data.setArrivalTime(++currentTime); //The sampling rate is 1 sec, hence to compute the currentTime we increment by 1 the time for each processed data point
            slideDataList.add(data);
            
            if(reportDataOutlierRuleBased){
               infDBH.reportOutlier(data, "ruleBased",ruleMeas); 
            }
        }

        if(slideDataList.size() > Constants.slide){
            notProcessedDataList.addAll(new ArrayList(slideDataList.subList(Constants.slide, slideDataList.size())));
            currentTime -= slideDataList.subList(Constants.slide, slideDataList.size()).size();
            slideDataList.subList(Constants.slide, slideDataList.size()).clear();
        }

        if(slideDataList.size() == Constants.slide){
            //Execute the MCOD algorithm
            mcodOutliers(mcod, slideDataList, currentTime);
            slideDataList.clear();
        }
    }
    
    private void mcodOutliers(MCOD mcnew, List<Data> slideDataList, long currentTime) {
        ArrayList<Data> outliers9 = mcnew.detectOutlier(slideDataList, currentTime, W, slide);

        //if (((double) mcnew.getDataListSize()) > k * 1.2) { //k*1.2 is a buffer for cold start
            for (Data outlier : outliers9) {
                double outlierLifeInSlides = 0;
                if (outlier.getOutlier_end_time() != 0) {
                    outlierLifeInSlides = (outlier.getOutlier_end_time() / slide) - (outlier.getOutlier_start_time() / slide) + 1;
                } else {
                    outlierLifeInSlides = (currentTime / slide) - (outlier.getOutlier_start_time() / slide) + 1;
                }
                if (outlierLifeInSlides >= ((double) W / slide) * outlierLife) {
                    if (outlier.getArrivalTime() > currentTime - W) {
                        infDBH.reportOutlier(outlier, "mcod", "");
                    }
                }
            }
        //}
    }
}
