/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.atl.smartmaintenance.faultdetection;

import com.atl.afaultdetection.utils.InfluxDBHandler;
import com.atl.mcod.MCOD;
import com.atl.afaultdetection.utils.Measurement;
import com.atl.afaultdetection.utils.Utils;
import com.atl.mcod.utils.Constants;
import com.atl.mcod.utils.Data;
import java.time.Instant;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 *
 * @author thanasis
 */
public class FDKnowledge {
    
    private MCOD mcod;
    private InfluxDBHandler infDBH;
    private double R;
    private int k;
    private int W;
    private int slide;
    private double outlierLife;
    private String ruleMeas;
    private Map<String, Measurement> measMap;
    long currentTime;
    
    //Holds the data of each slide
    //MCOD with knowledge is count-based not time based, hence the slideDataList should be processed only if it's size is equal to the slide size
    private List<Data> slideDataList;
    private List<Data> notProcessedDataList;

    public FDKnowledge(Map<String,Measurement> measMap, InfluxDBHandler infDBH, double R, int k, int W, int slide, double outlierLife) {
        this.R = R;
        this.k = k;
        this.W = W;
        this.slide = slide;
        this.outlierLife = outlierLife;
        this.mcod = new MCOD(R, k, W, slide);
        this.infDBH = infDBH;
        this.measMap = measMap;
        if (Constants.algorithms.contains("ruleBasedKnowledge")) {
            this.ruleMeas = Utils.checkRuleBasedParameters(measMap);
        }
        this.slideDataList = new ArrayList<Data>();
        this.notProcessedDataList = new ArrayList<Data>();
        this.currentTime = 0;
    }
    
    public void mcodWithKnowledge(Map<String,Iterator<Map.Entry<Instant,Double>>> slideDataIteratorsMap) {        
        if(slideDataIteratorsMap.size() < 2){
            return;
        }
        
        if(!notProcessedDataList.isEmpty()){
            if(notProcessedDataList.size() <= Constants.slide){
                slideDataList.addAll(notProcessedDataList);
                notProcessedDataList.clear();
                currentTime += slideDataList.size();
            } else {
                currentTime += Constants.slide;
                mcodOutliers(mcod, notProcessedDataList.subList(0, Constants.slide-1), currentTime);
                notProcessedDataList.subList(0, Constants.slide).clear();
                slideDataList.addAll(notProcessedDataList);
                notProcessedDataList.clear();
                currentTime += slideDataList.size();
            }
        }
        
        double previousStamperPosition = -1;
        int stabilizationOffset = Constants.stabilizationPeriod;
        
        //loop until one of the iterator runs out of data. 
        //Since we have used group by and fill(linear) in the query it is expected all the iterators to have the same amount of data.
        iteratorsLoop:
        while (Utils.allIteratorsHaveNext(slideDataIteratorsMap)) {
            Instant date = null;//it is used to mark the actual arrival date of the current data point. We have used precision of a second in the query so it is expected all the current points from different measurements to have the same timestamp

            boolean skipValues = false;
            Map<String, Map.Entry<Instant,Double>> valuesMap = new HashMap<>(); //holds the actual <time,point> pair for each different measurment
            for (Map.Entry<String, Iterator<Map.Entry<Instant,Double>>> entry : slideDataIteratorsMap.entrySet()) {
                String key = entry.getKey();
                Map.Entry<Instant,Double> value = entry.getValue().next();

                date = value.getKey();
//                System.out.println(date);
                //There might be a bug in the InfluxDB java library, as there are some null values returned that actually don't exist.
                //Skip the null values.
                if (value.getValue() == null) {
                    skipValues = true; //continue the outer loop
                }
                
                //If the stamper is locked (above a specific value) discard some values to be sure that the oil level is stabilized
                if (key.contains(Constants.measShortNameMap.get("stamperPosition"))) {
                    if((value.getValue() >= Constants.stamperLockedPosition && previousStamperPosition < Constants.stamperLockedPosition) || stabilizationOffset < Constants.stabilizationPeriod){
                        stabilizationOffset -= 1;
                        if(stabilizationOffset == 0){
                            stabilizationOffset = Constants.stabilizationPeriod;
                        } else {
                            skipValues = true; //continue the outer loop
                        }
                    }
                    previousStamperPosition = value.getValue();
                }

                //InfluxDB does not accept constraints in specific values of multi-measurement queries
                //Hence the exclusion of zero oil level values (which is possibly a sensor fault) should be done manually.
                if (value.getValue() == 0 && key.contains(Constants.measShortNameMap.get("oilLevel"))) {
                    skipValues = true; //break the outer loop
                }

                //Ignore the values that are less than zero possibly due to a sensor problem
                if (value.getValue() < 0) {
                    skipValues = true; //continue the outer loop
                }
                
                //Ignore products 12 and 3 as there is no clear stamper locked position for those
                if (key.contains(Constants.measShortNameMap.get("productNumber")) && (value.getValue() == 12 || value.getValue() == 3)) {
                    skipValues = true; //continue the outer loop
                }
                
                //Ignore values if the stamper is not locked (above a specific high)
                if (key.contains(Constants.measShortNameMap.get("stamperPosition")) && value.getValue() < Constants.stamperLockedPosition) {
                    skipValues = true; //continue the outer loop 
                }
                
                valuesMap.put(key, value);
            }
            if(skipValues){
                continue;
            }

            //a flag to report data that violate a rule based threshold
            boolean reportDataOutlierRuleBased = false;
            
            //The implementation of the MCOD expects all the data point values in an array format
            //Exclude stamperPosition and ProductNumber
            double[] d = new double[valuesMap.size()-2];
            int i = 0;
            for (String key : valuesMap.keySet()) {
                double value = valuesMap.get(key).getValue();
                if(key.contains(Constants.measShortNameMap.get("stamperPosition")) || key.contains(Constants.measShortNameMap.get("productNumber"))){
                    continue;
                }
                
                if(Constants.algorithms.contains("ruleBasedKnowledge") && key.equals(ruleMeas)){
                    reportDataOutlierRuleBased = Utils.checkRule("ruleBasedKnowledge", key, value, ruleMeas);
                }
                
                Measurement meas = measMap.get(key);
                d[i] = Utils.normalize(value, meas.getMinValue(), meas.getMaxValue());
                i++;
            }
            
           
            //Data class is needed by the current MCOD implementation
            Data data = new Data(d);
            data.setActualTime(date);
            data.setArrivalTime(++currentTime); //The sampling rate is 1 sec, hence to compute the currentTime we increment by 1 the time for each processed data point.
            slideDataList.add(data);
            if(reportDataOutlierRuleBased){
               infDBH.reportOutlier(data, "ruleBasedKnowledge",ruleMeas); 
            }
        }
        
        if(slideDataList.size() > Constants.slide){
            notProcessedDataList.addAll(new ArrayList(slideDataList.subList(Constants.slide, slideDataList.size())));
            currentTime -= slideDataList.subList(Constants.slide, slideDataList.size()).size();
            slideDataList.subList(Constants.slide, slideDataList.size()).clear();
        }

        if(slideDataList.size() == Constants.slide){
            //Execute the MCOD algorithm
            mcodOutliers(mcod, slideDataList, currentTime);
            slideDataList.clear();
        }
    }
    
    private void mcodOutliers(MCOD mcnew, List<Data> slideDataList, long currentTime) {
        ArrayList<Data> outliers9 = mcnew.detectOutlier(slideDataList, currentTime, W, slide);
        //if (((double) mcnew.getDataListSize()) > k * 1.2) { //k*1.2 is a buffer for cold start
            for (Data outlier : outliers9) {
                double outlierLifeInSlides = 0;
                if (outlier.getOutlier_end_time() != 0) {
                    outlierLifeInSlides = (outlier.getOutlier_end_time() / slide) - (outlier.getOutlier_start_time() / slide) + 1;
                } else {
                    outlierLifeInSlides = (currentTime / slide) - (outlier.getOutlier_start_time() / slide) + 1;
                }
                if (outlierLifeInSlides >= ((double) W / slide) * outlierLife) {
                    if (outlier.getArrivalTime() > currentTime - W) {
                        infDBH.reportOutlier(outlier, "mcodKnowledge", "");
                    }
                }
            }
        //}
    }
}
